#include<assert.h>

//! \file memory1.cpp
//    .
//      ,   .

#define PAGE_SIZE 4096	//!<  

#define PAGE_COUNT (PAGE_SIZE-sizeof(void*))/sizeof(T)	//!<    

#pragma pack(push,1)

//!  
template<typename T>class Allocator{
	//!  
	union item_st{
		T obj;
		item_st *next;
	};
	//!  
	struct bank_st{
		union {
			struct {
				char undef[PAGE_SIZE-sizeof(bank_st*)];
				bank_st *next;
			};
			item_st array[PAGE_COUNT];
		};
	};
	bank_st *bank_list;	//!<   
	item_st *free_list;	//!<   
	void NewBank();
public:
	Allocator();
	~Allocator();
	void Clear();
	T *New();
	void Free(T*);
};

#pragma pack(pop)


template<class T>Allocator<T>::Allocator()
{
	assert(sizeof(T)<PAGE_SIZE/3);
	bank_list=0;
	free_list=0;
}


template<class T>Allocator<T>::~Allocator()
{
	Clear();
}

//!  .       
template<class T>void Allocator<T>::Clear()
{
	bank_st *next;
	while(bank_list) {
		next=bank_list->next;
		delete bank_list;
		bank_list=next;
	}
}

//!      
template<class T>void Allocator<T>::NewBank()
{
	bank_st *bank=new bank_st;
	int n=PAGE_COUNT;
	while(n) {
		n--;
		Free(&bank->array[n].obj);
	}
	bank->next=bank_list;
	bank_list=bank;
}

//!   (     )
template<class T>void Allocator<T>::Free(T *obj)
{
	item_st *item=(item_st*)obj;
	item->next=free_list;
	free_list=item;
}

//!   
template<class T>T* Allocator<T>::New()
{
	if(!free_list) NewBank();
	T *res=&free_list->obj;
	free_list=free_list->next;
	return res;
}

int main()
{
// 
	Allocator<int> obj1;
	Allocator<double> obj2;
	int *y=obj1.New();
	obj1.Free(y);
	double *d=obj2.New();
	obj2.Free(d);
	return 0;
}
